""" Script for generating statistics from sensitive workload data.

It would produce aggregated data per group (privatized). And finally per day 
and hour statistics.

The idea of this script is to create a release of aggregated statistics 
which does not allow for unprivatizing of the gid (which for me is SLURM's
account).

This script will produce:

    (invariant, that is unanonymized) 
    1) per week submit time % of jobs
    2) per hour submit time % of jobs
    
    (anonymized)
    3) % of jobs submitted per group 
    
    per group
    
    4) runtime histogram
    5) allocated cpus histogram

Copyright (C) 2024
Manuel G. Marciani
BSC-CNS Earth Sciences Department
"""

# library to handle differential privacy
# in-house library for handling workload format
import pyswf
import os 
import argparse
import pdb

def invariant_statistics(swftrace: pyswf.SWFtrace, output_path: str) -> None:
    """ This function computes all the statistics that are not meant to be 
    altered by anonimization, hence called invariants.

    It will produce plots for the weekly and hourly submit time of the machine.
    Those will be saved to files.

    swftrace: the SWFtrace object with the workload data 
    output_path: root folder of the output release
    """
    # for timedelta calculations for each one of the timestamps 
    from datetime import timedelta

    # invariants (not privatized released): submit, start, end, uid, gid (account)
    # we are computing (in timestamp) each of the instants, for then joining per day of the week 
    swftrace.job_list["submit_ts"] = swftrace.job_list.submit_time.apply(lambda instant: timedelta(seconds=instant)) + swftrace.start_time
    swftrace.job_list["start_ts"] = swftrace.job_list.start_time.apply(lambda instant: timedelta(seconds=instant)) + swftrace.start_time
    swftrace.job_list["end_ts"] = swftrace.job_list.end_time.apply(lambda instant: timedelta(seconds=instant)) + swftrace.start_time

    # submit info
    # weekday will be 0, Monday, 1, Tuesday, 2 Wednesday, ...
    swftrace.job_list["submit_day_of_the_week"] = swftrace.job_list["submit_ts"].apply(lambda date: date.weekday())
    swftrace.job_list["submit_hour_of_the_day"] = swftrace.job_list["submit_ts"].apply(lambda date: date.hour)

    # aggregated counts
    no_jobs_submit_agg_per_day_of_the_week = swftrace.job_list.groupby("submit_day_of_the_week").count()["job_id"]
    no_jobs_submit_agg_per_hour_of_the_day = swftrace.job_list.groupby("submit_hour_of_the_day").count()["job_id"]

    with open(f"{output_path}/release/invariant/submit_agg_jobs_per_day_of_the_week.csv", "w") as output_submit_agg_per_day_of_the_week:
        for key, value in no_jobs_submit_agg_per_day_of_the_week.items():
             output_submit_agg_per_day_of_the_week.write(f"{key},{value}\n")

    with open(f"{output_path}/release/invariant/submit_agg_jobs_per_hour_of_the_day.csv", "w") as output_submit_agg_per_hour_of_the_day:
        for key, value in no_jobs_submit_agg_per_hour_of_the_day.items():
             output_submit_agg_per_hour_of_the_day.write(f"{key},{value}\n")

def variant_statistics(swftrace: pyswf.SWFtrace, max_influence: int, budget: tuple[float, float], output_dir: str) -> None:
    """ Create the anonymized statistics and output them both in graphs and the raw data

    swftrace: input standard workflow format object
    max_influence: this is the maximum changes that the privatized entity might cause on the (end) distribution
    budget: output distance assuming it is laplace :)
    output_dir: root folder of the output
    """

    import opendp.prelude as dp
    dp.enable_features("contrib", "floating-point")

    # assuming symmetric difference for the input dataset distance. which one is better(?)

    # these are the columns of each of the dataframes with the implicit column 
    # as well "df_id"
    col_names = ["df_id"] + list(swftrace.job_list.columns)
    
    #private: per group (slurm account) % of jobs from account (sum
    gids = list(swftrace.job_list.gid.unique())

    # this is pretty ugly, but works. We will convert the dataframe to csv
    # this is just a deterministic transformation.
    per_gid_num_jobs_histogram = (
        dp.t.make_split_dataframe(separator=",", col_names=col_names) >>
        dp.t.make_select_column(key="gid", TOA=str) >>
        # i64 might be too much! 
        dp.t.then_cast(TOA=dp.i64) >>
        dp.t.then_impute_constant(-1) >>
        dp.t.then_count_by_categories(categories=gids)
    )
    
    noisy_per_gid_num_jobs_histogram = dp.binary_search_chain(
        lambda s: per_gid_num_jobs_histogram >> dp.m.then_laplace(scale=s),
        d_in=max_influence, d_out=budget[0]
    )
   
    # output number of jobs per group id
    with open(f"{output_dir}/release/variant/per_gid_num_jobs.csv", "w") as output_per_gid_num_jobs:
        for key, value in zip(gids,noisy_per_gid_num_jobs_histogram(swftrace.job_list.to_csv(header=False))):
            output_per_gid_num_jobs.write(f"{key},{value}\n")        

    #private: per group (slurm account) runtime histogram (in percentage). bounded to 0 and max(runtime)
    #private: per group (slurm account) allocated_cpus (in percentage) histogram bounded to 0 and max_procs

    # taken from https://docs.opendp.org/en/stable/getting-started/examples/histograms.html
    def make_laplace_threshold_budget(
        preprocess: dp.Transformation,
        d_in, d_out
    ) -> dp.Measurement:
        """Make a stability histogram that respects a given d_in, d_out."""
        def privatize(s, t = 1e8):
            return preprocess >> dp.m.then_laplace_threshold(scale=s, threshold=t)
        
        s = dp.binary_search_param(lambda s: privatize(s = s), d_in, d_out)
        t = dp.binary_search_param(lambda t: privatize(s = s, t = t), d_in, d_out)
    
        return privatize(s=s, t=t)

    # transform for runtime and allocated cpus
    preprocess_alloc_cpus_runtime = (
        dp.t.make_split_dataframe(separator=",", col_names=col_names) >>
        dp.t.make_select_column(key="runtime",TOA=str) >>
        dp.t.then_cast(TOA=dp.i64) >>
        dp.t.then_count_by(MO=dp.L1Distance[float], TV=float)
    )
    
    noisy_histogram = make_laplace_threshold_budget(
        preprocess_alloc_cpus_runtime,
        d_in=max_influence,
        d_out=budget   
    )        

    for gid in gids:
        # sensitive group's info 
        gid_jobs = swftrace.job_list[swftrace.job_list.gid == gid]
        with open(f"{output_dir}/release/variant/{gid}_runtime.csv", "w") as gid_runtime_output:
            gid_runtime_output.write("runtime,num_observations\n")
            for key, value in noisy_histogram(gid_jobs.to_csv(header=False)).items():
                gid_runtime_output.write(f"{key},{round(value)}\n")

        with open(f"{output_dir}/release/variant/{gid}_alloc_cpus.csv", "w") as gid_alloc_cpus_output:
            gid_alloc_cpus_output.write("alloc_cpus,num_observations\n")
            for key, value in noisy_histogram(gid_jobs.to_csv(header=False)).items():
                gid_alloc_cpus_output.write(f"{key},{round(value)}\n")                

def arg_parse() -> argparse.ArgumentParser:
    """
    Parse arguments from the command line and check correctness.
    Folder structure will be created if not present 
    :returns: argparse object with the parameters
    """
    parser = argparse.ArgumentParser(description="Anonymize using Differential Privacy workloads which are in Standard Workflow format")
    parser.add_argument("swf_path", type=str, help="relative or absolute path to the swf file")
    parser.add_argument("out_dir", type=str, help="relative or absolute path to the output directory")

    arguments = parser.parse_args()

    import os, sys

    arguments.swf_path = os.path.abspath(arguments.swf_path)
    if not os.path.exists(arguments.swf_path):
        sys.exit(f'opendp_release.py: file {arguments.swf_path} does not exist')

    arguments.out_dir = os.path.abspath(arguments.out_dir)
    if not os.path.exists(f"{arguments.out_dir}/release"):
        print(f'opendp_release.py: output folder {arguments.out_dir} does not exist. Creating it...')
        os.mkdir(f"{arguments.out_dir}/release") 
        os.mkdir(f"{arguments.out_dir}/release/invariant") 
        os.mkdir(f"{arguments.out_dir}/release/variant") 

    return arguments

if __name__ == "__main__":
    args = arg_parse()    

    # load data
    workload = pyswf.SWFtrace.from_swf(args.swf_path)
    invariant_statistics(workload, args.out_dir)
    variant_statistics(workload, 1, (1., 1e-8), args.out_dir)
   
